#include "picoled/SSD1306.h"

#include <cstring>

// https://cdn-shop.adafruit.com/datasheets/SSD1306.pdf
namespace
{

constexpr uint8_t MEMORY_MODE    = 0x20;
constexpr uint8_t COLUMN_ADDRESS = 0x21;
constexpr uint8_t PAGE_ADDRESS   = 0x22;
constexpr uint8_t CONTRAST       = 0x81;
constexpr uint8_t CHARGE_PUMP    = 0x8D;
constexpr uint8_t SEGMENT_REMAP  = 0xA0;
constexpr uint8_t ALLON_RESUME   = 0xA4;
constexpr uint8_t ALLON          = 0xA5;
constexpr uint8_t NORMAL_DISPLAY = 0xA6;
constexpr uint8_t INVERT_DISPLAY = 0xA7;
constexpr uint8_t MULTIPLEX      = 0xA8;
constexpr uint8_t OFF            = 0xAE;
constexpr uint8_t ON             = 0xAF;
constexpr uint8_t COMSCAN_INC    = 0xC0;
constexpr uint8_t COMSCAN_DEC    = 0xC8;
constexpr uint8_t OFFSET         = 0xD3;
constexpr uint8_t CLKDIV         = 0xD5;
constexpr uint8_t PRECHARGE      = 0xD9;
constexpr uint8_t COMPINS        = 0xDA;
constexpr uint8_t VCOM_DETECT    = 0xDB;
constexpr uint8_t LOW_COLUMN     = 0x00;
constexpr uint8_t HIGH_COLUMN    = 0x10;
constexpr uint8_t STARTLINE      = 0x40;
constexpr uint8_t RIGHT_HSCROLL  = 0x26;
constexpr uint8_t LEFT_HSCROLL   = 0x27;
constexpr uint8_t VRIGHT_HSCROLL = 0x29;
constexpr uint8_t VLEFT_HSCROLL  = 0x2A;
constexpr uint8_t SCROLL_OFF     = 0x2E;
constexpr uint8_t SCROLL_ON      = 0x2F;
constexpr uint8_t VSCROLL_AREA   = 0xA3;

}

namespace picoled
{

ssd1306::ssd1306(i2c_inst* i2c, uint8_t address) :
  oled(128, 64),
  i2c(i2c),
  address(address),
  buffer(new uint8_t[128 * 64 / 8 + 1])
{
  // the user is expected to set GPIO func and pullup
  // as well as initialize the i2c instance

  command(OFF);
  command(CLKDIV); command(0x80);
  command(MULTIPLEX); command(63);
  command(OFFSET); command(0);
  command(STARTLINE | 0);
  command(MEMORY_MODE); command(0);
  command(SEGMENT_REMAP | 0);
  command(COMSCAN_DEC);
  command(COMPINS); command(0x12);
  command(CONTRAST); command(0xCF);
  command(PRECHARGE); command(0x22);
  command(VCOM_DETECT); command(0x40);
  command(NORMAL_DISPLAY);
  command(SCROLL_OFF);
  command(CHARGE_PUMP); command(0x14);
  command(ALLON_RESUME);
  command(ON);

  buffer[0] = 0x40;
}

ssd1306::~ssd1306()
{
  delete[] buffer;
}

void ssd1306::update() const
{
  // clear display to zero
  memset(&buffer[1], 0, width * height * sizeof(*buffer));

  for (int x = 0; x < width; x++)
  {
    for (int y = 0; y < height; y++)
    {
      if (!(*this)(x, y).is_on()) { continue; }

      const int index = (width - x - 1) + (y / 8) * width;
      const uint8_t bits = 1 << (y & 7);

      // add one to leave room for 0x40 data indicator at beginning of array
      buffer[index + 1] |= bits;
    }
  }

  command(PAGE_ADDRESS); command(0); command(7);
  command(COLUMN_ADDRESS); command(0); command(width - 1);
  i2c_write_blocking(i2c, address, buffer, width * height / 8 + 1, false);
}

void ssd1306::command(uint8_t cmd) const
{
  const uint8_t values[2] = { 0x00, cmd };
  i2c_write_blocking(i2c, address, values, 2, false);
}

}
